;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; classes.lisp -- persistent objects via metaobjects
;;; 
;;; Initial version 8/26/2004 by Andrew Blumberg
;;; <ablumberg@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(in-package "ELEPHANT")

(defmethod initialize-instance :before  ((instance persistent)
					 &rest initargs
					 &key from-oid
					 spec 
					 ;; Putting the default use
					 ;; of the global variable here 
					 ;; is very bad for testing and multi-repository
					 ;; use; it is, however, good for making
					 ;; things work exactly the way they originally did!
					 (sc *store-controller*))
  "Sets the OID."
  (declare (ignore initargs))

;; This lines are fundamentally valuable in making sure that 
;; we hvae completely specified things.
;;  (if (null sc)
;;      (break))
  (if (not from-oid)
      (setf (oid instance) (next-oid sc))
      (setf (oid instance) from-oid))
  (if (not spec)
      (if (not (typep sc 'bdb-store-controller))
	  (setf (:dbcn-spc-pst instance) (:dbcn-spc sc))
	  (setf (:dbcn-spc-pst instance) (controller-path sc))
	  )
      (setf (:dbcn-spc-pst instance) spec))
  (cache-instance sc instance))

(defclass persistent-object (persistent)
  ()
  (:documentation 
   "Superclass of all user-defined persistent classes.  This is
automatically inherited if you use the persistent-metaclass
metaclass.")
  (:metaclass persistent-metaclass))

(defmethod shared-initialize :around ((class persistent-metaclass) slot-names &rest args &key direct-superclasses)
  "Ensures we inherit from persistent-object."
  (let* ((persistent-metaclass (find-class 'persistent-metaclass))
	 (persistent-object (find-class 'persistent-object))
	 (not-already-persistent (loop for superclass in direct-superclasses
				       never (eq (class-of superclass) persistent-metaclass))))
    (if (and (not (eq class persistent-object)) not-already-persistent)
	(apply #'call-next-method class slot-names
	       :direct-superclasses (cons persistent-object
					  direct-superclasses) args)
	(call-next-method))))

#+allegro
(defun make-persistent-reader (name slot-definition class class-name)
  (eval `(defmethod ,name ((instance ,class-name))
	  (slot-value-using-class ,class instance ,slot-definition))))

#+allegro
(defun make-persistent-writer (name slot-definition class class-name)
  (eval `(defmethod (setf ,name) ((instance ,class-name) value)
	  (setf (slot-value-using-class ,class instance ,slot-definition)
	   value))))

#+allegro
(defmethod initialize-accessors ((slot-definition persistent-slot-definition) class)
  (let ((readers (slot-definition-readers slot-definition))
	(writers (slot-definition-writers slot-definition))
	(class-name (class-name class)))
    (loop for reader in readers
	  do (make-persistent-reader reader slot-definition class class-name))
    (loop for writer in writers
	  do (make-persistent-writer writer slot-definition class class-name))))

#+allegro
(defmethod reinitialize-instance :around ((instance persistent-metaclass) &rest initargs &key &allow-other-keys)
  (prog1
      (call-next-method)
    (when (class-finalized-p instance)
      (update-persistent-slots instance (persistent-slot-names instance))
      (loop with persistent-slots = (persistent-slots instance)
	    for slot-def in (class-direct-slots instance)
	    when (member (slot-definition-name slot-def) persistent-slots)
	    do (initialize-accessors slot-def instance))
      (make-instances-obsolete instance))))

#+(or cmu sbcl)
(defmethod reinitialize-instance :around ((instance persistent-metaclass) &rest initargs &key &allow-other-keys)
  (prog1
      (call-next-method)
    (when (class-finalized-p instance)
      (update-persistent-slots instance (persistent-slot-names instance))
      (make-instances-obsolete instance))))

#+allegro
(defmethod finalize-inheritance :around ((instance persistent-metaclass))
  (prog1
      (call-next-method)
    (if (not (slot-boundp instance '%persistent-slots))
	(setf (%persistent-slots instance) 
	      (cons (persistent-slot-names instance) nil)))))

#+(or cmu sbcl)
(defmethod finalize-inheritance :around ((instance persistent-metaclass))
  (prog1
      (call-next-method)
    (if (not (slot-boundp instance '%persistent-slots))
	(setf (%persistent-slots instance) 
	      (cons (persistent-slot-names instance) nil)))))

(defmethod shared-initialize :around ((instance persistent-object) slot-names &rest initargs &key &allow-other-keys)
  "Initializes the persistent slots via initargs or forms.
This seems to be necessary because it is typical for
implementations to optimize setting the slots via initforms
and initargs in such a way that slot-value-using-class et al
aren't used.  Calls the next method for the transient slots."
  (let* ((class (class-of instance))
	 (persistent-slot-names (persistent-slot-names class)))
    (flet ((persistent-slot-p (item) 
	     (member item persistent-slot-names :test #'eq)))
      (let ((transient-slot-inits 
	     (if (eq slot-names t)	; t means all slots
		 (transient-slot-names class)
		 (remove-if #'persistent-slot-p slot-names)))
	    (persistent-slot-inits
	     (if (eq slot-names t) persistent-slot-names
		 (remove-if-not #'persistent-slot-p slot-names))))
	;; initialize the persistent slots
	(flet ((initialize-from-initarg (slot-def)
		 (loop for initarg in initargs
		    with slot-initargs = (slot-definition-initargs slot-def)
		    when (member initarg slot-initargs :test #'eq)
		    do 
		    (setf (slot-value-using-class class instance slot-def) 
			  (getf initargs initarg))
		    (return t))))
	  (loop for slot-def in (class-slots class)
	     unless 
	     (initialize-from-initarg slot-def)
	     when 
	     (member (slot-definition-name slot-def) persistent-slot-inits :test #'eq)
	     unless 
	     (slot-boundp-using-class class instance slot-def)
	     do
	     (let ((initfun (slot-definition-initfunction slot-def)))
	       (when initfun
		 (setf (slot-value-using-class class instance slot-def)
		       (funcall initfun))))
	     )
	  ;; let the implementation initialize the transient slots
	  (apply #'call-next-method instance transient-slot-inits initargs))))))

(defmethod update-instance-for-redefined-class :around ((instance persistent-object) added-slots discarded-slots property-list &rest initargs &key &allow-other-keys)
  ;; probably should delete discarded slots, but we'll worry about that later
  (prog1
      (call-next-method)
    (let* ((class (class-of instance))
	   (new-persistent-slots (set-difference (persistent-slots class)
						 (old-persistent-slots class))))
	   
      (apply #'shared-initialize instance new-persistent-slots initargs))))

(defun find-slot-def-by-name (class slot-name)
  (loop for slot-def in (class-slots class)
	when (eq (slot-definition-name slot-def) slot-name)
	do (return slot-def)))

(defmethod update-instance-for-different-class :around ((previous persistent) (current persistent) &rest initargs &key)
  (let* ((old-class (class-of previous))
	 (new-class (class-of current))
	 (new-persistent-slots (set-difference
				(persistent-slots new-class)
				(persistent-slots old-class)))
	 (raw-retained-persistent-slots (intersection (persistent-slots new-class)
						      (persistent-slots old-class)))
	 (retained-unbound-slots (loop for slot-name in raw-retained-persistent-slots
				       when (not (persistent-slot-boundp previous slot-name))
				       collect slot-name))
	 (retained-persistent-slots (set-difference raw-retained-persistent-slots retained-unbound-slots)))
    (apply #'shared-initialize current (append new-persistent-slots retained-unbound-slots) initargs)
    (loop for slot-def in (class-slots new-class)
	  when (member (slot-definition-name slot-def) retained-persistent-slots)
	  do (setf (slot-value-using-class new-class
					   current
					   slot-def)
		   (slot-value-using-class old-class
					   previous
					   (find-slot-def-by-name old-class (slot-definition-name slot-def)))))
    (call-next-method)))

(defmethod slot-value-using-class :around ((class persistent-metaclass) (instance persistent-object) (slot-def persistent-slot-definition))
  "Get the slot value from the database."
  (declare (optimize (speed 3)))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-reader instance name)))

(defmethod (setf slot-value-using-class) :around (new-value (class persistent-metaclass) (instance persistent-object) (slot-def persistent-slot-definition))
  "Set the slot value in the database."
  (declare (optimize (speed 3)))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-writer new-value instance name)))

(defmethod slot-boundp-using-class :around ((class persistent-metaclass) (instance persistent-object) (slot-def persistent-slot-definition))
  "Checks if the slot exists in the database."
  (declare (optimize (speed 3)))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-boundp instance name)))

(defmethod slot-boundp-using-class :around ((class persistent-metaclass) (instance persistent-object) (slot-name symbol))
  "Checks if the slot exists in the database."
  (declare (optimize (speed 3)))
  (loop for slot in (class-slots class)
	for matches-p = (eq (slot-definition-name slot) slot-name)
	until matches-p
	finally (if (and matches-p
			 (typep slot 'persistent-slot-definition))    
		    (persistent-slot-boundp instance slot-name)
		    (call-next-method))))

(defmethod slot-makunbound-using-class :around ((class persistent-metaclass) (instance persistent-object) (slot-def persistent-slot-definition))
  "Deletes the slot from the database."
  (declare (optimize (speed 3))
  	   (ignore class))
   (if (sql-store-spec-p (:dbcn-spc-pst instance))
       (progn
 	(let* ((sc (check-con (:dbcn-spc-pst instance)))
 	       (con (controller-db sc)))
 	(sql-remove-from-root
 	 (form-slot-key (oid instance) (slot-definition-name slot-def))
 	  sc
 	  con
 	 )
 	))
       (with-buffer-streams (key-buf)
 	(buffer-write-int (oid instance) key-buf)
 	(serialize (slot-definition-name slot-def) key-buf)
 	(db-delete-buffered
 	 (controller-db (check-con (:dbcn-spc-pst instance))) key-buf
 	 :transaction *current-transaction*
 	 :auto-commit *auto-commit*))
       )
  instance)

#+allegro
(defmethod slot-makunbound-using-class :around ((class persistent-metaclass) (instance persistent-object) (slot-name symbol))
  (loop for slot in (class-slots class)
	until (eq (slot-definition-name slot) slot-name)
	finally (if (typep slot 'persistent-slot-definition)
		    (slot-makunbound-using-class class instance slot)
		    (call-next-method))))
