;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; controller.lisp -- Lisp interface to a Berkeley DB store
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")


;; This list contains functions that take one arugment,
;; the "spec", and will construct an appropriate store
;; controller from it.
(defvar *strategies* '())

(defvar *elephant-lib-path* "/usr/local/share/common-lisp/elephant-0.3/")

(defun register-strategy (spec-to-controller)
  (setq *strategies* (delete spec-to-controller *strategies*))
  (setq *strategies* (cons spec-to-controller *strategies*))
  )

(defun get-controller (spec)
  (let ((store-controllers nil))
    (dolist (s *strategies*)
      (let ((sc (funcall s spec)))
	(if sc
	    (push sc store-controllers))))
    (if (not (= (length store-controllers) 1))
	(error "Strategy resolution for this spec completely failed!")
	(car store-controllers))
    ))


(defclass store-controller ()  
  ;; purely abstract class doesn't need a slot, though it 
  ;; should take the common ones.
  ((path :type (or pathname string)
	 :accessor controller-path
	 :initarg :path)
    (root :reader controller-root)
    (db :type (or null pointer-void) :accessor controller-db :initform '())
   (environment :type (or null pointer-void) 
		:accessor controller-environment)
   (oid-db :type (or null pointer-void) :accessor controller-oid-db)
   (oid-seq :type (or null pointer-void) :accessor controller-oid-seq)
   (btrees :type (or null pointer-void) :accessor controller-btrees)
   (indices :type (or null pointer-void) :accessor controller-indices)
   (indices-assoc :type (or null pointer-void) 
		  :accessor controller-indices-assoc)
   (instance-cache :accessor instance-cache
		   :initform (make-cache-table :test 'eql)))
  (:documentation "Class of objects responsible for the
book-keeping of holding DB handles, the cache, table
creation, counters, locks, the root (for garbage collection,)
et cetera."))

(defclass bdb-store-controller (store-controller)  
  (
   )
  (:documentation "Class of objects responsible for the
book-keeping of holding DB handles, the cache, table
creation, counters, locks, the root (for garbage collection,)
et cetera."))

;; Without somemore sophistication, these functions 
;; need to be defined here, so that they will be available for testing
;; even if you do not use the strategy in question...
(defun bdb-store-spec-p (path)
  (stringp path))

(defun sql-store-spec-p (path)
  (listp path))


;; This has now way of passing in optionals?
(defun bdb-test-and-construct (spec)
  (if (bdb-store-spec-p spec)
      (open-store-bdb spec)
      nil)
  )

(eval-when ( :load-toplevel)
  (register-strategy 'bdb-test-and-construct)
  )

(defgeneric open-controller (sc &key recover recover-fatal thread)
  (:documentation 
   "Opens the underlying environment and all the necessary
database tables."))

(defgeneric close-controller (sc)
  (:documentation 
   "Close the db handles and environment.  Tries to wipe out
references to the db handles."))

(defgeneric build-btree (sc)
  (:documentation 
   "Construct a btree of the appropriate type corresponding to this store-controller."))
 
(defgeneric build-indexed-btree (sc)
  (:documentation 
   "Construct a btree of the appropriate type corresponding to this store-controller."))

(defgeneric get-transaction-macro-symbol (sc)
  (:documentation 
   "Return the strategy-specific macro symbol that will let you do a transaction within that macro."))


(defun make-indexed-btree (&optional (sc *store-controller*))
  (build-indexed-btree sc)
  )


(defgeneric build-btree-index (sc &key primary key-form)
  (:documentation 
   "Construct a btree of the appropriate type corresponding to this store-controller."))

(defgeneric copy-from-key (key src dst)
  (:documentation 
   "Move the object identified by key on the root in the src to the dst."))

(defmethod copy-from-key (key src dst)
  (let ((v (get-from-root key :store-controller src)))
    (if v
	(add-to-root key v :store-controller dst)
	v))
  )

(defun copy-btree-contents (src dst)
  (map-btree 
   #'(lambda (k v)
       (setf (get-value k dst) v)
       )
   src)
  )

;; I don't know if I need a "deeper" copy here or not....
(defun my-copy-hash-table (ht)
  (let ((nht (make-hash-table)))
    (maphash
     #'(lambda (k v) 
	 (setf (gethash k nht) v))
     ht)
    nht)
  )
 
(defun add-index-from-index (iname v dstibt dstsc)
  (declare (type btree-index v)
	   (type indexed-btree dstibt))
  (let ((kf (key-form v)))
    (format t " kf ~A ~%" kf)
    (let ((index
	   (build-btree-index dstsc :primary dstibt
			      :key-form kf)))
      ;; Why do I have to do this here?
      (setf (indices dstibt) (make-hash-table))
      (setf (indices-cache dstibt) (make-hash-table))
      (setf (gethash iname (indices-cache dstibt)) index)
      (setf (gethash iname (indices dstibt)) index)
      )
    )
  )

(defun my-copy-indices (ht dst dstsc)
  (maphash
   #'(lambda (k v) 
       (add-index-from-index k v dst dstsc))
   ht)
  )
 
(defmethod migrate ((dst store-controller) obj)
  "Copy a currently persistent object to a new repository."
  (if (typep obj 'btree)
      ;; For a btree, we need to copy the object with the indices intact,
      ;; then just read it out...
      (if (typep obj 'indexed-btree) 
	  ;; We have to copy the indexes..
 	  (let ((nobj (build-indexed-btree dst)))
 	    (my-copy-indices (indices obj) nobj dst)
 	    (copy-btree-contents obj nobj)
 	    nobj
 	    )
 	  (let ((nobj (build-btree dst)))
 	    (copy-btree-contents obj nobj)
 	    nobj)
 	  )
      (error (format nil "the migrate function cannot migrate objects like ~A~%" obj)
	     )))
 
;; ;; This routine attempst to do a destructive migration
;; ;; of the object to the new repository
(defmethod migraten-pobj ((dst store-controller) obj copy-fn) 
   "Migrate a persistent object and apply a binary (lambda (dst src) ...) function to the new object."
   ;; The simplest thing to do here is to make 
   ;; an object of the new class;
   ;; we will make it the responsibility of the caller to 
   ;; perform the copy on the slots --- or 
   ;; we can force them to pass in this function.
   (if (typep obj 'persistent)
       (let ((nobj (make-instance (type-of obj) :sc dst)))
  	(apply copy-fn (list nobj obj))
  	nobj)
       (error (format "obj ~A is not a persistent object!~%" obj))
       )
   )
 

(defun add-to-root (key value &key (store-controller *store-controller*))
  "Add an arbitrary persistent thing to the root, so you can
retrieve it in a later session.  N.B. this means it (and
everything it points to) won't get gc'd."
  (declare (type store-controller store-controller))
  (setf (get-value key (controller-root store-controller)) value))

(defun get-from-root (key &key (store-controller *store-controller*))
  "Get a something from the root."
  (declare (type store-controller store-controller))
  (get-value key (controller-root store-controller)))

(defun from-root-existsp (key &key (store-controller *store-controller*))
  "Get a something from the root."
  (declare (type store-controller store-controller))
  (if (existsp key (controller-root store-controller))
      t 
      nil))

(defun remove-from-root (key &key (store-controller *store-controller*))
  "Remove something from the root."
  (declare (type store-controller store-controller))
  (remove-kv key (controller-root store-controller)))

(defun cache-instance (sc obj)
  "Cache a persistent object with the controller."
  (declare (type store-controller sc))
  (setf (get-cache (oid obj) (instance-cache sc)) obj))

(defun get-cached-instance (sc oid class-name)
  "Get a cached instance, or instantiate!"
  (declare (type store-controller sc)
	   (type fixnum oid))
  (let ((obj (get-cache oid (instance-cache sc))))
    (if obj obj
	;; Should get cached since make-instance calls cache-instance
	(make-instance class-name :from-oid oid))))

(defmethod next-oid ((sc bdb-store-controller))
  "Get the next OID."
  (declare (type bdb-store-controller sc))
  (db-sequence-get-fixnum (controller-oid-seq sc) 1 :transaction +NULL-VOID+
			  :auto-commit t :txn-nosync t))

;; Open/close     
(defmethod open-controller ((sc bdb-store-controller) &key (recover nil)
			    (recover-fatal nil) (thread t))
  (let ((env (db-env-create)))
    ;; thread stuff?
    (setf (controller-environment sc) env)
    (db-env-open env (controller-path sc) :create t :init-txn t :init-lock t 
		 :init-mpool t :init-log t :thread thread
		 :recover recover :recover-fatal recover-fatal)
    (let ((db (db-create env))
	  (btrees (db-create env))
	  (indices (db-create env))
	  (indices-assoc (db-create env)))
      (setf (controller-db sc) db)
      (setf (gethash (controller-path sc) *dbconnection-spec*) sc)
      (db-open db :file "%ELEPHANT" :database "%ELEPHANTDB" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-btrees sc) btrees)
      (sleepycat::db-set-lisp-compare btrees)
      (db-open btrees :file "%ELEPHANT" :database "%ELEPHANTBTREES" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-indices sc) indices)
      (sleepycat::db-set-lisp-compare indices)
      (sleepycat::db-set-lisp-dup-compare indices)
      (db-set-flags indices :dup-sort t)
      (db-open indices :file "%ELEPHANT" :database "%ELEPHANTINDICES" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-indices-assoc sc) indices-assoc)
      (sleepycat::db-set-lisp-compare indices-assoc)
      (sleepycat::db-set-lisp-dup-compare indices-assoc)
      (db-set-flags indices-assoc :dup-sort t)
      (db-open indices-assoc :file "%ELEPHANT" :database "%ELEPHANTINDICES" 
	       :auto-commit t :type DB-UNKNOWN :thread thread :rdonly t)
      (sleepycat::db-fake-associate btrees indices-assoc :auto-commit t)
      
      (let ((db (db-create env)))
	(setf (controller-oid-db sc) db)
	(db-open db :file "%ELEPHANTOID" :database "%ELEPHANTOID" 
		 :auto-commit t :type DB-BTREE :create t :thread thread)
	(let ((oid-seq (db-sequence-create db)))
	  (db-sequence-set-cachesize oid-seq *cachesize*)
	  (db-sequence-set-flags oid-seq :seq-inc t :seq-wrap t)
	  (db-sequence-set-range oid-seq 0 most-positive-fixnum)
	  (db-sequence-initial-value oid-seq 0)
	  (db-sequence-open oid-seq "%ELEPHANTOID"
			    :auto-commit t :create t :thread t)
	  (setf (controller-oid-seq sc) oid-seq)))

      (let ((root (make-instance 'bdb-btree :from-oid -1 :sc sc)))
	(setf (slot-value sc 'root) root))
      sc)))

(defmethod close-controller ((sc bdb-store-controller))
  (when (slot-value sc 'root)
    ;; no root
    (setf (slot-value sc 'root) nil)
    ;; clean instance cache
    (setf (instance-cache sc) (make-cache-table :test 'eql))
    ;; close handles / environment
    (db-sequence-close (controller-oid-seq sc))
    (setf (controller-oid-seq sc) nil)
    (db-close (controller-oid-db sc))
    (setf (controller-oid-db sc) nil)
    (db-close (controller-indices-assoc sc))
    (setf (controller-indices-assoc sc) nil)
    (db-close (controller-indices sc))
    (setf (controller-indices sc) nil)
    (db-close (controller-btrees sc))
    (setf (controller-btrees sc) nil)
    (db-close (controller-db sc))
    (setf (controller-db sc) nil)
    (db-env-close (controller-environment sc))
    (setf (controller-environment sc) nil)
    nil))

;; Do these things need to take &rest arguments?
(defmethod build-btree ((sc bdb-store-controller))
  (make-bdb-btree sc)
  )


(defun make-btree (&optional (sc *store-controller*))
  (build-btree sc)
  )

(defmethod get-transaction-macro-symbol ((sc bdb-store-controller))
  'with-transaction
  )

(defun open-store (spec  &key (recover nil)
  		   (recover-fatal nil) (thread t))
    "Conveniently open a store controller."
  (setq *store-controller*  
	(get-controller spec))
  (open-controller *store-controller* :recover recover 
		   :recover-fatal recover-fatal :thread thread))

(defun open-store-bdb (spec  &key (recover nil)
		       (recover-fatal nil) (thread t))
  "Conveniently open a store controller."
  (setq *store-controller*  
 	(if (bdb-store-spec-p spec)
	    (make-instance 'bdb-store-controller :path spec)
	    (error (format nil "uninterpretable path/spec specifier: ~A" spec))))
  (open-controller *store-controller* :recover recover 
		   :recover-fatal recover-fatal :thread thread))


(defmacro with-open-store-bdb ((path) &body body)
  "Executes the body with an open controller,
 unconditionally closing the controller on exit."
  `(let ((*store-controller* (make-instance 'bdb-store-controller :path ,path)))
     (declare (special *store-controller*))
     (open-controller *store-controller*)
     (unwind-protect
	  (progn ,@body)
       (close-controller *store-controller*))))

(defmacro with-open-controller ((&optional (sc '*store-controller*))
				&body body)
  "Executes body with the specified controller open, closing
the controller unconditionally on exit."
  `(unwind-protect
       (progn
	 (let (*store-controller* (open-controller ,sc))
	   (declare (special *store-controller*))
	   ,@body))
     (close-controller ,sc)))

(defun close-store ()
  "Conveniently close the store controller."
  (if *store-controller*
  (close-controller *store-controller*)))

(defmacro with-open-store ((spec) &body body)
  "Executes the body with an open controller,
unconditionally closing the controller on exit."
  `(let ((*store-controller* 
	  (get-controller ,spec)))
     (declare (special *store-controller*))
;;     (open-controller *store-controller*)
     (unwind-protect
	  (progn ,@body)
       (close-controller *store-controller*))))


;;; Make these respect the transaction keywords (e.g. degree-2)
;; (defun start-transaction (&key (parent *current-transaction*))
;;   "Start a transaction.  May be nested but not interleaved."
;;   (vector-push-extend *current-transaction* *transaction-stack*)
;;   (setq *current-transaction* 
;; 	(db-transaction-begin (controller-environment *store-controller*)
;; 			      :parent parent)))

(defun start-ele-transaction (&key (parent *current-transaction*) (store-controller *store-controller*))
    "Start a transaction.  May be nested but not interleaved."
    (vector-push-extend *current-transaction* *transaction-stack*)
    (setq *current-transaction* 
	(db-transaction-begin (controller-environment store-controller)
  			      :parent parent)))

(defun commit-transaction ()
  "Commit the current transaction."
  (db-transaction-commit)
  (setq *current-transaction* (vector-pop *transaction-stack*)))

(defun abort-transaction ()
  "Abort the current transaction."
  (db-transaction-abort)
  (setq *current-transaction* (vector-pop *transaction-stack*)))

(defgeneric persistent-slot-reader-aux (sc instance name)
  (:documentation 
   "Auxilliary method to allow implementation-specific slot reading"))

(defgeneric persistent-slot-writer-aux (sc new-value instance name)
  (:documentation 
   "Auxilliary method to allow implementation-specific slot writing"))

