;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; metaclasses.lisp -- persistent objects via metaobjects
;;; 
;;; Initial version 8/26/2004 by Andrew Blumberg
;;; <ablumberg@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

(defvar *dbconnection-spec* 
  (make-hash-table :test 'equal))

(defun connection-is-indeed-open (con)
  t ;; I don't yet know how to implement this
  )

;; This needs to be a store-controller method...
(defun check-con (spec &optional sc )
  (let ((con (gethash spec *dbconnection-spec*)))
    (if (and con (connection-is-indeed-open con))
	con
	(if (not (typep sc 'bdb-store-controller))
 	    (progn
	      (error "We can't default to *store-controller* in a multi-use enviroment."))
	    ;; 	    (setf (gethash spec *dbconnection-spec*)
	    ;; 		  (clsql:connect (cdr (:dbcn-spc sc))
	    ;; 				 :database-type :postgresql-socket
	    ;; 				 :if-exists :old)))
	    (error "We don't know how to open a bdb-connection here!")
	    ;; if they don't give us connection-spec, we can't reopen things...
	    ))))



(defclass persistent ()
  ((%oid :accessor oid :initarg :from-oid)
  ;; This is just an idea for storing connections in the persistent
  ;; objects; these should be transient as well, if that flag exists!
  ;; In the case of sleepy cat, this is the controller-db from 
  ;; the store-controller.  In the case of SQL this is
  ;; the connection spec (since the connection might be broken?)
  ;; It probably would be better to put a string in here in the case
  ;; of sleepycat...
  (dbonnection-spec-pst :type (or list string) :accessor :dbcn-spc-pst :initarg :dbconnection-spec-pst
;; Andrew Philpot discovered that this breaks the ACL implementation.
;; We are green under both SBCL and ACL with it commented out, so I am commiting it. -- rlr 06 JAN 2006
;;			:initform '()
)
   )
  (:documentation 
   "Abstract superclass for all persistent classes (common
to user-defined classes and collections.)"))

(defclass persistent-metaclass (standard-class)
  ((%persistent-slots :accessor %persistent-slots))
  (:documentation 
   "Metaclass for persistent classes.  Use this metaclass to
define persistent classes.  All slots are persistent by
default; use the :transient flag otherwise."))

(defmethod persistent-slots ((class persistent-metaclass))
  (car (%persistent-slots class)))

(defmethod persistent-slots ((class standard-class))
  nil)

(defmethod old-persistent-slots ((class persistent-metaclass))
  (cdr (%persistent-slots class)))

(defmethod update-persistent-slots ((class persistent-metaclass) new-slot-list)
;;    (setf (%persistent-slots class) (cons new-slot-list (car (%persistent-slots class)))))
   (setf (%persistent-slots class) (cons new-slot-list 
					 (if (slot-boundp class '%persistent-slots)
 					    (car (%persistent-slots class))
					    nil)
 					    )))

(defclass persistent-slot-definition (standard-slot-definition)
  ())

(defclass persistent-direct-slot-definition (standard-direct-slot-definition persistent-slot-definition)
  ())

(defclass persistent-effective-slot-definition (standard-effective-slot-definition persistent-slot-definition)
  ())

(defclass transient-slot-definition (standard-slot-definition)
  ((transient :initform t :initarg :transient :allocation :class)))

(defclass transient-direct-slot-definition (standard-direct-slot-definition transient-slot-definition)
  ())

(defclass transient-effective-slot-definition (standard-effective-slot-definition transient-slot-definition)
  ())

(defgeneric transient (slot))

(defmethod transient ((slot standard-direct-slot-definition))
  t)

(defmethod transient ((slot persistent-direct-slot-definition))
  nil)

#+allegro
(defmethod excl::valid-slot-allocation-list ((class persistent-metaclass))
  '(:instance :class :database))

(defmethod slot-definition-allocation ((slot-definition persistent-slot-definition))
  :database)

(defmethod direct-slot-definition-class ((class persistent-metaclass) &rest initargs)
  "Checks for the transient tag (and the allocation type)
and chooses persistent or transient slot definitions."
  (let ((allocation-key (getf initargs :allocation))
	(transient-p (getf initargs :transient)))
    (when (consp transient-p) (setq transient-p (car transient-p)))
    (cond ((and (eq allocation-key :class) transient-p)
	   (find-class 'transient-direct-slot-definition))
	  ((and (eq allocation-key :class) (not transient-p))
	   (error "Persistent class slots are not supported, try :transient t."))
	  (transient-p
	   (find-class 'transient-direct-slot-definition))
	  (t
	   (find-class 'persistent-direct-slot-definition)))))

(defmethod validate-superclass ((class persistent-metaclass) (super standard-class))
  "Persistent classes may inherit from ordinary classes."
  t)

(defmethod validate-superclass ((class standard-class) (super persistent-metaclass))
  "Ordinary classes may NOT inherit from persistent classes."
  nil)

(defgeneric persistent-p (class))

(defmethod persistent-p ((class t))
  nil)

(defmethod persistent-p ((class persistent-metaclass))
  t)

(defmethod persistent-p ((class persistent-slot-definition))
  t)

(defmethod effective-slot-definition-class ((class persistent-metaclass) &rest initargs)
  "Chooses the persistent or transient effective slot
definition class depending on the keyword."
  (let ((transient-p (getf initargs :transient)))
    (when (consp transient-p) (setq transient-p (car transient-p)))
    (cond (transient-p
	   (find-class 'transient-effective-slot-definition))
	  (t
	   (find-class 'persistent-effective-slot-definition)))))

#+openmcl
(defmethod compute-effective-slot-definition ((class persistent-metaclass) slot-name direct-slot-definitions)
  (declare (ignore slot-name))
  (apply #'make-effective-slot-definition class
	 (compute-effective-slot-definition-initargs 
	  class direct-slot-definitions)))

#+openmcl
(defmethod compute-effective-slot-definition-initargs ((class slots-class)
						       direct-slots)
  (let* ((name (loop for s in direct-slots
		  when s
		  do (return (slot-definition-name s))))
	 (initer (dolist (s direct-slots)
                   (when (%slot-definition-initfunction s)
                     (return s))))
         (documentor (dolist (s direct-slots)
                       (when (%slot-definition-documentation s)
                         (return s))))
         (first (car direct-slots))
         (initargs (let* ((initargs nil))
                     (dolist (dslot direct-slots initargs)
                       (dolist (dslot-arg (%slot-definition-initargs  dslot))
                         (pushnew dslot-arg initargs :test #'eq))))))
    (list
     :name name
     :allocation (%slot-definition-allocation first)
     :documentation (when documentor (nth-value
                                      1
                                      (%slot-definition-documentation
                                       documentor)))
     :class (%slot-definition-class first)
     :initargs initargs
     :initfunction (if initer (%slot-definition-initfunction initer))
     :initform (if initer (%slot-definition-initform initer))
     :type (or (%slot-definition-type first) t))))

(defun ensure-transient-chain (slot-definitions initargs)
  (declare (ignore initargs))
  (loop for slot-definition in slot-definitions
     always (transient slot-definition)))

(defmethod compute-effective-slot-definition-initargs ((class persistent-metaclass) slot-definitions)
  (let ((initargs (call-next-method)))
    (if (ensure-transient-chain slot-definitions initargs)
	(append initargs '(:transient t))
	(progn
	  (setf (getf initargs :allocation) :database)
	  initargs))))


(defmacro persistent-slot-reader (instance name)
`(if (not (bdb-store-spec-p  (:dbcn-spc-pst ,instance)))
       (persistent-slot-reader-aux (check-con (:dbcn-spc-pst ,instance)) ,instance ,name)
       (progn
	 (with-buffer-streams (key-buf value-buf)
	   (buffer-write-int (oid ,instance) key-buf)
	   (serialize ,name key-buf)
	   (let ((buf (db-get-key-buffered 
		       (controller-db (check-con (:dbcn-spc-pst ,instance))) 
						 key-buf value-buf)))
		   (if buf (deserialize buf  :sc (check-con (:dbcn-spc-pst instance)))
		       #+cmu
		       (error 'unbound-slot :instance ,instance :slot ,name)
		       #-cmu
		       (error 'unbound-slot :instance ,instance :name ,name)))))))

#+(or cmu sbcl)
(defun make-persistent-reader (name)
  (lambda (instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-reader instance name)))

(defmacro persistent-slot-writer (new-value instance name)
  `(if (not (bdb-store-spec-p (:dbcn-spc-pst ,instance)))
       (persistent-slot-writer-aux (check-con (:dbcn-spc-pst ,instance)) ,new-value ,instance ,name)
       (with-buffer-streams (key-buf value-buf)
	 (buffer-write-int (oid ,instance) key-buf)
	 (serialize ,name key-buf)
	 (serialize ,new-value value-buf)
	 (db-put-buffered 
	  (controller-db (check-con (:dbcn-spc-pst ,instance)))
	  key-buf value-buf
	  :transaction *current-transaction*
	  :auto-commit *auto-commit*)
	 ,new-value)))

#+(or cmu sbcl)
(defun make-persistent-writer (name)
  (lambda (new-value instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-writer new-value instance name)))

;; This this is not a good way to form a key...
(defun form-slot-key (oid name)
  (format nil "~A ~A" oid name)
  )

(defmacro persistent-slot-boundp (instance name)
  `(if (not (bdb-store-spec-p (:dbcn-spc-pst ,instance)))
       (persistent-slot-boundp-aux (check-con (:dbcn-spc-pst ,instance)) ,instance ,name)
       (progn
	 (with-buffer-streams (key-buf value-buf)
	   (buffer-write-int (oid ,instance) key-buf)
	   (serialize ,name key-buf)
	   (let ((buf (db-get-key-buffered 
		       (controller-db (check-con (:dbcn-spc-pst ,instance)))
		       key-buf value-buf)))
	     (if buf T nil))))))

#+(or cmu sbcl)
(defun make-persistent-slot-boundp (name)
  (lambda (instance)
    (declare (optimize (speed 3))
	     (type persistent-object instance))
    (persistent-slot-boundp instance name)))

#+(or cmu sbcl)
(defmethod initialize-internal-slot-functions ((slot-def persistent-slot-definition))
  (let ((name (slot-definition-name slot-def)))
    (setf (slot-definition-reader-function slot-def)
	  (make-persistent-reader name))
    (setf (slot-definition-writer-function slot-def)
	  (make-persistent-writer name))
    (setf (slot-definition-boundp-function slot-def)
	  (make-persistent-slot-boundp name)))
  slot-def)

(defun persistent-slot-names (class)
  (let ((slot-definitions (class-slots class)))
    (loop for slot-definition in slot-definitions
       when (equalp (class-of slot-definition) (find-class 'persistent-effective-slot-definition))
       collect (slot-definition-name slot-definition))))

(defun transient-slot-names (class)
  (let ((slot-definitions (class-slots class)))
    (loop for slot-definition in slot-definitions
       unless (persistent-p slot-definition)
       collect (slot-definition-name slot-definition))))
