;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; utils.lisp -- utility functions
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(in-package "ELEPHANT")

(declaim (type fixnum *lisp-obj-id*)
	 (type hash-table *circularity-hash*)
	 (type boolean *auto-commit*))

(defvar *cachesize* 100
  "Size of the OID sequence cache.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Thread-local specials

(defparameter *store-controller* nil 
  "The store controller which persistent objects talk to.")

;; Specials which control persistent objects
(defvar *auto-commit* T
  "Commit things not in transactions?")

(defvar *transaction-stack* (make-array 0 :adjustable t :fill-pointer t)
  "Used if the user manually creates transactions.")

;; Stuff the serializer uses
(defvar *lisp-obj-id* 0 
  "Circularity ids for the serializer.")
(defvar *circularity-hash* (make-hash-table :test 'eq)
  "Circularity hash for the serializer.")

#+(or cmu sbcl allegro)
(defvar *resourced-byte-spec* (byte 32 0)
  "Byte specs on CMUCL, SBCL and Allegro are conses.")

;; TODO: make this for real!
(defun run-elephant-thread (thunk)
  "Sets the specials (which hopefully are thread-local) to
make the Elephant thread-safe."
  (let ((*current-transaction* +NULL-VOID+)
	(sleepycat::*errno-buffer* (allocate-foreign-object :int 1))
	;; if vector-push-extend et al are thread-safe, this
	;; doesn't need to be thread-local.
	(sleepycat::*buffer-streams* 
	 (make-array 0 :adjustable t :fill-pointer t))
	(*store-controller* *store-controller*)
	(*auto-commit* *auto-commit*)
	(*transaction-stack*
	 (make-array 0 :adjustable t :fill-pointer t))
	(*lisp-obj-id* 0)
	(*circularity-hash* (make-hash-table :test 'eq))
	#+(or cmu sbcl allegro)
	(*resourced-byte-spec* (byte 32 0)))
    (declare (special *current-transaction* sleepycat::*errno-buffer*
		      sleepycat::*buffer-streams*
		      *store-controller* *auto-commit* *transaction-stack*
		      *lisp-obj-id* *circularity-hash* 
		      #+(or cmu sbcl allegro) *resourced-byte-spec*))
    (funcall thunk)))

;; get rid of spot idx and adjust the arrray
(defun remove-indexed-element-and-adjust (idx array)
  (let ((last (- (length array) 1)))
    (do ((i idx (1+ i)))
	((= i last) nil)
      (progn
	(setf (aref array i) (aref array (+ 1 i)))))
    (adjust-array array last)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Macros
;; Good defaults for elephant
(defmacro with-transaction (
 			    (&key transaction 
 				  (store-controller '*store-controller*)
 				  environment 
 				  (parent '*current-transaction*)
 				  degree-2 dirty-read txn-nosync
 				  txn-nowait txn-sync
 				  (retries 100))
			    &body body
)
  "Execute a body with a transaction in place.  On success,
the transaction is committed.  Otherwise, the transaction is
aborted.  If the body deadlocks, the body is re-executed in
a new transaction, retrying a fixed number of iterations.
*auto-commit* is false for the body of the transaction."
  `(if (not (typep ,store-controller 'elephant::bdb-store-controller))
       (elephant::with-transaction-sql (:store-controller-sql ,store-controller)
	 ,@body)
;;        (if (clsql::in-transaction-p 
;;  	    :database 
;;  	    (controller-db ,store-controller))
;;  	   (progn
;;  	     ,@body)
;;  	   (prog2
;;  	       (clsql::set-autocommit nil)
;;  	       (clsql::with-transaction
;;  		   (:database 
;;  		    (controller-db ,store-controller))
;;  		 ,@body)
;;  	     (clsql::set-autocommit t)))
       (let ((env (if ,environment ,environment 
 		      (controller-environment ,store-controller))))
 	       (sleepycat:with-transaction (:transaction ,transaction
 							 :environment env
 							 :parent ,parent
							 :degree-2 ,degree-2
 							 :dirty-read ,dirty-read
 							 :txn-nosync ,txn-nosync
 							 :txn-nowait ,txn-nowait
 							 :txn-sync ,txn-sync
 							 :retries ,retries)
 
 		 (let ((*auto-commit* nil))
 		   ,@body)))
 	     ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Portable value-weak hash-tables for the cache: when the
;;; values are collected, the entries (keys) should be
;;; flushed from the table too

(defun make-cache-table (&rest args)
  "Make a values-weak hash table: when a value has been
collected, so are the keys."
  #+(or cmu sbcl scl)
  (apply #'make-hash-table args)
  #+allegro
  (apply #'make-hash-table :values :weak args)
  #+lispworks
  (apply #'make-hash-table :weak-kind :value args)
  #+openmcl
  (apply #'make-hash-table :weak :value args)
  #-(or cmu sbcl scl allegro lispworks)
  (apply #'make-hash-table args)
  )

#+openmcl
(defclass cleanup-wrapper ()
  ((cleanup :accessor cleanup :initarg :cleanup)
   (value :accessor value :initarg :value)))

#+openmcl
(defmethod ccl:terminate ((c cleanup-wrapper))
  (funcall (cleanup c)))

(defun get-cache (key cache)
  "Get a value from a cache-table."
  #+(or cmu sbcl)
  (let ((val (gethash key cache)))
    (if val (values (weak-pointer-value val) t)
	(values nil nil)))
  #+openmcl 
  (let ((wrap (gethash key cache)))
    (if wrap (values (value wrap) t)
	(values nil nil)))
  #+(or allegro lispworks)
  (gethash key cache)
  )

(defun make-finalizer (key cache)
  #+(or cmu sbcl)
  (lambda () (remhash key cache))
  #+(or allegro openmcl)
  (lambda (obj) (declare (ignore obj)) (remhash key cache))
  )

(defun setf-cache (key cache value)
  "Set a value in a cache-table."
  #+(or cmu sbcl)
  (let ((w (make-weak-pointer value)))
    (finalize value (make-finalizer key cache))
    (setf (gethash key cache) w)
    value)
  #+openmcl
  (let ((w (make-instance 'cleanup-wrapper :value value
			  :cleanup (make-finalizer key cache))))
    (ccl:terminate-when-unreachable w)
    (setf (gethash key cache) w)
    value)
  #+allegro
  (progn
    (excl:schedule-finalization value (make-finalizer key cache))
    (setf (gethash key cache) value))
  #+lispworks
  (setf (gethash key cache) value)
  )

(defsetf get-cache setf-cache)

